# -*- coding: utf-8 -*-
from .extension_tables import ExtensionTablesLCA
from .one_spatial_scale import OneSpatialScaleLCA
from .two_spatial_scales import TwoSpatialScalesLCA
from .two_spatial_scales_weighting import TwoSpatialScalesWithGenericLoadingLCA
